from typing import Any

import torch

from turbo_alignment.dataset.chat.models import ChatDatasetRecord
from turbo_alignment.generators.base import ChatGeneratorBase
from turbo_alignment.settings.generators.outputs.chat import (
    AnswerMessage,
    ChatInferenceOutput,
)


class ChatGenerator(ChatGeneratorBase[ChatDatasetRecord, ChatInferenceOutput]):
    def _generate_from_batch_records(
        self,
        records: list[dict[str, torch.Tensor]],
        original_records: list[ChatDatasetRecord],
        dataset_name: str,
    ) -> list[ChatInferenceOutput]:
        input_ids = [record['input_ids'].tolist() for record in records]
        attention_mask = [record['attention_mask'].tolist() for record in records]

        max_input_length = max(len(sample) for sample in input_ids)

        batch = self._tokenizer.pad(
            {
                'input_ids': input_ids,
                'attention_mask': attention_mask,
            },
            padding='max_length',
            max_length=max_input_length,
            return_tensors='pt',
        )

        batched_input_ids = batch['input_ids'].to(self.device)
        batched_attention_mask = batch['attention_mask'].to(self.device)

        output_indices = self._model.generate(
            inputs=batched_input_ids,
            attention_mask=batched_attention_mask,
            generation_config=self._transformers_generator_parameters,
            tokenizer=self._tokenizer,
            pad_token_id=self._tokenizer.pad_token_id,
        )

        postprocessed_output_indices = self._postprocess(
            input_indices=batched_input_ids,
            output_indices=output_indices,
            remove_prompt=self._custom_generation_settings.remove_prompt,
        )

        answers = self._decode(token_indices=postprocessed_output_indices)

        return [
            ChatInferenceOutput(
                id=original_record.id,
                dataset_name=dataset_name,
                messages=original_record.messages,
                label=original_record.label,
                meta=original_record.meta,
                answers=[
                    AnswerMessage(
                        id='0',
                        content=answer,
                        input_token_ids=None,
                        answer_token_ids=None,
                        logits=None,
                    )
                ],
            )
            for original_record, answer in zip(original_records, answers)
        ]

    def _generate_from_single_record(
        self,
        record: dict[str, Any],
        original_record: ChatDatasetRecord,
        dataset_name: str,
    ) -> ChatInferenceOutput:
        input_ids = torch.unsqueeze(record['input_ids'], 0).to(self.device)
        attention_mask = torch.unsqueeze(record['attention_mask'], 0).to(self.device)

        output_indices = self._model.generate(
            inputs=input_ids,
            attention_mask=attention_mask,
            generation_config=self._transformers_generator_parameters,
            tokenizer=self._tokenizer,
            pad_token_id=self._tokenizer.pad_token_id,
        )

        postprocessed_output_indices = self._postprocess(
            input_indices=input_ids,
            output_indices=output_indices,
            remove_prompt=self._custom_generation_settings.remove_prompt,
        )

        answers = self._decode(token_indices=postprocessed_output_indices)

        logits: torch.Tensor | None = None
        input_token_ids: torch.Tensor | None = None
        answer_tokens_ids: torch.Tensor | None = None

        if self._return_logits:
            with torch.no_grad():
                logits = self._model(output_indices).logits.cpu()

            answer_tokens_ids = postprocessed_output_indices
            input_token_ids = input_ids

            answer_messages = [
                AnswerMessage(
                    id=str(i),
                    content=a,
                    input_token_ids=input_token_ids,
                    answer_token_ids=a_t_ids.unsqueeze(0),
                    logits=l.unsqueeze(0),
                )
                for i, (a, a_t_ids, l) in enumerate(zip(answers, answer_tokens_ids, logits))  # type: ignore[arg-type]
            ]
        else:
            answer_messages = [
                AnswerMessage(
                    id=str(i),
                    content=a,
                    input_token_ids=input_token_ids,
                    answer_token_ids=answer_tokens_ids,
                    logits=logits,
                )
                for i, a in enumerate(answers)
            ]

        return ChatInferenceOutput(
            id=original_record.id,
            dataset_name=dataset_name,
            messages=original_record.messages,
            label=original_record.label,
            meta=original_record.meta,
            answers=answer_messages,
        )



from pathlib import Path

from pydantic_settings import BaseSettings

from turbo_alignment.common import set_random_seed
from turbo_alignment.settings.cherry_pick import CherryPickSettings
from turbo_alignment.settings.datasets.base import MultiDatasetSettings
from turbo_alignment.settings.logging.clearml import ClearMLSettings
from turbo_alignment.settings.logging.weights_and_biases import WandbSettings
from turbo_alignment.settings.model import (
    ModelForPeftSettings,
    PreTrainedAdaptersModelSettings,
    PreTrainedModelSettings,
)
from turbo_alignment.settings.s3 import CheckpointUploaderCallbackParameters
from turbo_alignment.settings.tf.special_tokens_setter import SpecialTokensSettings
from turbo_alignment.settings.tf.tokenizer import TokenizerSettings
from turbo_alignment.settings.tf.trainer import TrainerSettings


class EarlyStoppingSettings(BaseSettings):
    patience: int = 1
    threshold: float | None = 0.0


class BaseTrainExperimentSettings(BaseSettings):
    log_path: Path = Path('train_output')
    seed: int = 42

    trainer_settings: TrainerSettings

    tokenizer_settings: TokenizerSettings
    special_tokens_settings: SpecialTokensSettings

    model_settings: (ModelForPeftSettings | PreTrainedModelSettings | PreTrainedAdaptersModelSettings)

    train_dataset_settings: MultiDatasetSettings
    val_dataset_settings: MultiDatasetSettings

    logging_settings: (WandbSettings | ClearMLSettings)

    checkpoint_uploader_callback_parameters: CheckpointUploaderCallbackParameters | None = None
    cherry_pick_settings: CherryPickSettings | None = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.log_path.mkdir(exist_ok=True)
        set_random_seed(self.seed)



from typing import Iterable

from accelerate import Accelerator
from transformers import PreTrainedModel, PreTrainedTokenizerBase

from turbo_alignment.cherry_picks.base import CherryPickCallbackBase
from turbo_alignment.dataset.pair_preferences import PairPreferenceDataset
from turbo_alignment.generators.rm import RMPairGenerator
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.settings.cherry_pick import RMCherryPickSettings
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults


class RmCherryPickCallback(CherryPickCallbackBase[PairPreferenceDataset]):
    def __init__(
        self,
        cherry_pick_settings: RMCherryPickSettings,
        datasets: Iterable[PairPreferenceDataset],
        metrics: list[Metric],
    ) -> None:
        super().__init__(cherry_pick_settings=cherry_pick_settings, datasets=datasets, metrics=metrics)

    def _get_dataset_metrics(
        self,
        dataset: PairPreferenceDataset,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> list[MetricResults]:
        accelerator: Accelerator = kwargs.get('accelerator', None)

        generator = RMPairGenerator(
            model=model,
            tokenizer=tokenizer,
            accelerator=accelerator,
        )

        generations = generator.generate_from_dataset(dataset)
        generations_w = [gen.reward_w for gen in generations]
        generations_l = [gen.reward_l for gen in generations]

        pair_scores = [1 if w > l else 0 for w, l in zip(generations_w, generations_l)]

        answers_w = [record.answer_w.content for record in generations]
        answers_l = [record.answer_l.content for record in generations]

        metric_outputs = [
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'chosen', values=answers_w)]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'rejected', values=answers_l)
                ]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'chosen_reward', values=generations_w)
                ]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'rejected_reward', values=generations_l)
                ]
            ),
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'score', values=pair_scores)]
            ),
        ]

        return metric_outputs



from collections import defaultdict
from typing import Any

import numpy as np
from scipy.stats import entropy
from transformers import PreTrainedTokenizerBase

from turbo_alignment.metrics.metric import Metric
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.DIVERSITY)
class DiversityMetric(Metric):
    def compute(self, **kwargs) -> list[MetricResults]:
        tokenizer: PreTrainedTokenizerBase = kwargs.get('tokenizer', None)
        predictions: list[list[str]] = kwargs.get('predictions', None)
        dataset_name: str = kwargs.get('dataset_name', '')

        if predictions is None:
            raise ValueError('predictions should not be None')

        if tokenizer is None:
            raise ValueError('tokenizer should not be None')

        element_wise_diversity_scores = [
            ElementWiseScores(
                label=dataset_name + '@@' + 'diversity',
                values=[self.average_token_entropy(answer_group, tokenizer) for answer_group in predictions],
            )
        ]

        return [
            MetricResults(element_wise_scores=element_wise_diversity_scores, need_average=need_average)
            for need_average in self._settings.need_average
        ]

    def average_token_entropy(self, answer_group: list[str], tokenizer: PreTrainedTokenizerBase) -> float:
        entropies = [self.token_entropy(answer, tokenizer) for answer in answer_group]
        if entropies:
            return sum(entropies) / len(entropies)

        return np.nan

    @staticmethod
    def token_entropy(sample: str, tokenizer: PreTrainedTokenizerBase) -> float:
        stats: dict[int, Any] = defaultdict(int)
        num_tokens = 0
        tokens = tokenizer.encode(sample)
        for t in tokens:
            if t == tokenizer.pad_token_id:
                continue
            stats[t] += 1
            num_tokens += 1
        for k in stats.keys():
            stats[k] /= num_tokens

        return entropy(list(stats.values()))



from typing import Any

from wandb.sdk.lib.disabled import RunDisabled
from wandb.sdk.wandb_run import Run

import wandb
from turbo_alignment.settings.logging.weights_and_biases import WandbSettings


def create_wandb_run(parameters: WandbSettings, config: dict[str, Any] | None = None) -> Run | RunDisabled:
    wandb_run = wandb.init(
        project=parameters.project_name,
        name=parameters.run_name,
        entity=parameters.entity,
        notes=parameters.notes,
        tags=parameters.tags,
        config=config,
        mode=parameters.mode,
    )

    if wandb_run is None:
        raise ValueError('Wandb run is None')

    return wandb_run



from transformers import PretrainedConfig, PreTrainedModel

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.modeling.liger_kernels.cross_entropy import LigerCrossEntropyLoss
from turbo_alignment.modeling.liger_kernels.geglu import LigerGEGLUMLP
from turbo_alignment.modeling.liger_kernels.rope import liger_rotary_pos_emb

logger = get_project_logger()


def apply_liger_kernel_to_gemma2(
    rope: bool = True,
    cross_entropy: bool = True,
    geglu: bool = True,
    model: PreTrainedModel = None,
) -> None:
    logger.info('Loading Liger-Kernels for Gemma2...')

    from transformers.models.gemma2 import modeling_gemma2

    if rope:
        modeling_gemma2.apply_rotary_pos_emb = liger_rotary_pos_emb
    if cross_entropy:
        modeling_gemma2.CrossEntropyLoss = LigerCrossEntropyLoss
    if geglu:
        modeling_gemma2.Gemma2MLP = LigerGEGLUMLP

    if model is not None:
        config: PretrainedConfig = model.config

        if hasattr(model, 'model'):
            base_model = model.model
        else:
            base_model = model

        torch_dtype = config.torch_dtype

        for decoder_layer in base_model.layers:
            if geglu:
                decoder_layer.mlp = LigerGEGLUMLP(config).to(torch_dtype)

    logger.info('Liger-Kernels have been successfully applied!')



from typing import Generator

import torch
from accelerate import Accelerator
from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.tf.loaders import load_model, load_tokenizer
from turbo_alignment.generators.multimodal import MultimodalGenerator
from turbo_alignment.modeling.multimodal.lm.projection import (
    ProjectionMultiModalModeling,
)
from turbo_alignment.pipelines.inference.base import BaseInferenceStrategy
from turbo_alignment.pipelines.mixin.multimodal import MultimodalMixin
from turbo_alignment.settings.pipelines.inference.multimodal import (
    MultimodalInferenceExperimentSettings,
    MultimodalSingleModelInferenceSettings,
)


class MultimodalInferenceStrategy(MultimodalMixin, BaseInferenceStrategy[MultimodalInferenceExperimentSettings]):
    @staticmethod
    def _load_model(
        inference_settings: MultimodalSingleModelInferenceSettings,
        tokenizer: PreTrainedTokenizerBase,
    ):
        language_model = load_model(inference_settings.model_settings, tokenizer)

        modality_encoders = MultimodalInferenceStrategy._load_modality_encoders(
            inference_settings.modality_encoder_settings_mapping,
            device=language_model.device,
            dtype=language_model.dtype,
        )

        model = ProjectionMultiModalModeling(
            language_model=language_model,
            encoders=modality_encoders,
            n_modality_embs=inference_settings.model_settings.n_modality_embeddings,
            modality_projector_mapping=inference_settings.modality_projector_mapping,
            modality_projector_initialization_mapping=None,
            peft=True,
        )
        model.modality_adapters.load_state_dict(torch.load(inference_settings.model_settings.projections_path))

        return model

    def _get_single_inference_settings(
        self, experiment_settings: MultimodalInferenceExperimentSettings, accelerator: Accelerator
    ) -> Generator[tuple[PreTrainedTokenizerBase, MultimodalGenerator, str, dict], None, None]:
        save_file_id = 0

        for model_inference_settings in experiment_settings.inference_settings:
            tokenizer = load_tokenizer(
                model_inference_settings.tokenizer_settings,
                model_inference_settings.model_settings,
            )
            model = self._load_model(model_inference_settings, tokenizer)

            for generation_settings in model_inference_settings.generation_settings:
                generator = MultimodalGenerator(
                    transformers_settings=generation_settings.transformers_settings,
                    custom_generation_settings=generation_settings.custom_settings,
                    tokenizer=tokenizer,
                    model=model,
                    accelerator=accelerator,
                )

                parameters_to_save = {
                    'model_settings': model_inference_settings.model_settings.dict(),
                    'generation_settings': generation_settings.dict(),
                }

                yield tokenizer, generator, f'single_inference_{save_file_id}.jsonl', parameters_to_save



from turbo_alignment.settings.pipelines.common import MergeAdaptersToBaseModelSettings
from turbo_alignment.settings.pipelines.sampling import (
    BaseSamplingWithRMSettings,
    RandomSamplingSettings,
    RSOSamplingSettings,
)
from turbo_alignment.settings.pipelines.train.classification import (
    ClassificationTrainExperimentSettings,
)
from turbo_alignment.settings.pipelines.train.ddpo import DDPOTrainExperimentSettings
from turbo_alignment.settings.pipelines.train.dpo import DPOTrainExperimentSettings
from turbo_alignment.settings.pipelines.train.kto import KTOTrainExperimentSettings
from turbo_alignment.settings.pipelines.train.multimodal import (
    MultimodalTrainExperimentSettings,
)
from turbo_alignment.settings.pipelines.train.rag import RAGTrainExperimentSettings
from turbo_alignment.settings.pipelines.train.rm import RMTrainExperimentSettings
from turbo_alignment.settings.pipelines.train.sft import SftTrainExperimentSettings



from typing import Generator

import torch
from accelerate import Accelerator
from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.tf.loaders import load_model, load_tokenizer
from turbo_alignment.generators.base import BaseGenerator
from turbo_alignment.generators.chat import ChatGenerator
from turbo_alignment.pipelines.inference.base import BaseInferenceStrategy
from turbo_alignment.settings.model import PreTrainedAdaptersModelSettings
from turbo_alignment.settings.pipelines.inference.chat import (
    ChatInferenceExperimentSettings,
)


class ChatInferenceStrategy(BaseInferenceStrategy[ChatInferenceExperimentSettings]):
    def _get_single_inference_settings(
        self, experiment_settings: ChatInferenceExperimentSettings, accelerator: Accelerator
    ) -> Generator[tuple[PreTrainedTokenizerBase, BaseGenerator, str, dict], None, None]:
        save_file_id = 0

        for model_inference_settings in experiment_settings.inference_settings:
            tokenizer = load_tokenizer(
                model_inference_settings.tokenizer_settings,
                model_inference_settings.model_settings,
            )

            if model_inference_settings.use_vllm:
                import vllm
                from vllm.lora.request import LoRARequest

                from turbo_alignment.generators.vllm_chat import VLLMChatGenerator

                lora_request: LoRARequest | None = None
                enable_lora: bool = False

                if isinstance(model_inference_settings.model_settings, PreTrainedAdaptersModelSettings):
                    lora_request = LoRARequest('adapter', 1, str(model_inference_settings.model_settings.adapter_path))
                    enable_lora = True

                model = vllm.LLM(
                    model=model_inference_settings.model_settings.model_path.absolute().as_posix(),
                    dtype='bfloat16',
                    tensor_parallel_size=model_inference_settings.tensor_parallel_size,
                    enable_lora=enable_lora,
                )

            else:
                model = load_model(model_inference_settings.model_settings, tokenizer)
                model = (
                    accelerator.prepare_model(model, device_placement=True, evaluation_mode=True)
                    if torch.cuda.is_available()
                    else model.to('cpu')  # type: ignore[attr-defined]
                )

            for generation_settings in model_inference_settings.generation_settings:
                generator_kwargs = {
                    'model': model,
                    'tokenizer': tokenizer,
                    'transformers_settings': generation_settings.transformers_settings,
                    'custom_generation_settings': generation_settings.custom_settings,
                    'batch': model_inference_settings.batch,
                }
                generator = (
                    ChatGenerator(
                        **generator_kwargs,
                        accelerator=accelerator,
                    )
                    if not model_inference_settings.use_vllm
                    else VLLMChatGenerator(
                        **generator_kwargs,
                        lora_request=lora_request,
                    )
                )

                parameters_to_save = {
                    'model_settings': model_inference_settings.model_settings.dict(),
                    'generation_settings': generation_settings.dict(),
                }

                save_file_id += 1

                yield tokenizer, generator, f'single_inference_{save_file_id}.jsonl', parameters_to_save



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['turbo_alignment',
 'turbo_alignment.cherry_picks',
 'turbo_alignment.cli',
 'turbo_alignment.common',
 'turbo_alignment.common.data',
 'turbo_alignment.common.data.multimodal',
 'turbo_alignment.common.data.multimodal.audio',
 'turbo_alignment.common.data.multimodal.image',
 'turbo_alignment.common.logging',
 'turbo_alignment.common.registry',
 'turbo_alignment.common.s3',
 'turbo_alignment.common.tf',
 'turbo_alignment.common.tf.callbacks',
 'turbo_alignment.common.tf.loaders',
 'turbo_alignment.common.tf.loaders.model',
 'turbo_alignment.dataset',
 'turbo_alignment.dataset.base',
 'turbo_alignment.dataset.chat',
 'turbo_alignment.dataset.classification',
 'turbo_alignment.dataset.ddpo',
 'turbo_alignment.dataset.kto',
 'turbo_alignment.dataset.multimodal',
 'turbo_alignment.dataset.pair_preferences',
 'turbo_alignment.dataset.sampling',
 'turbo_alignment.generators',
 'turbo_alignment.metrics',
 'turbo_alignment.modeling',
 'turbo_alignment.modeling.common',
 'turbo_alignment.modeling.imagebind',
 'turbo_alignment.modeling.imagebind.heads',
 'turbo_alignment.modeling.imagebind.postprocessors',
 'turbo_alignment.modeling.imagebind.preprocessors',
 'turbo_alignment.modeling.imagebind.trunks',
 'turbo_alignment.modeling.liger_kernels',
 'turbo_alignment.modeling.multimodal',
 'turbo_alignment.modeling.multimodal.encoders',
 'turbo_alignment.modeling.multimodal.encoders.audio',
 'turbo_alignment.modeling.multimodal.encoders.image',
 'turbo_alignment.modeling.multimodal.lm',
 'turbo_alignment.modeling.multimodal.projectors',
 'turbo_alignment.modeling.rag',
 'turbo_alignment.pipelines',
 'turbo_alignment.pipelines.inference',
 'turbo_alignment.pipelines.mixin',
 'turbo_alignment.pipelines.preprocessing',
 'turbo_alignment.pipelines.sampling',
 'turbo_alignment.pipelines.train',
 'turbo_alignment.settings',
 'turbo_alignment.settings.datasets',
 'turbo_alignment.settings.generators',
 'turbo_alignment.settings.generators.outputs',
 'turbo_alignment.settings.logging',
 'turbo_alignment.settings.pipelines',
 'turbo_alignment.settings.pipelines.common',
 'turbo_alignment.settings.pipelines.inference',
 'turbo_alignment.settings.pipelines.sampling',
 'turbo_alignment.settings.pipelines.train',
 'turbo_alignment.settings.tf',
 'turbo_alignment.trainers']

package_data = \
{'': ['*']}

install_requires = \
['albumentations>=1.3.1,<2.0.0',
 'bitsandbytes>=0.41.1,<0.42.0',
 'cached-path>=1.6.3,<2.0.0',
 'clearml>=1.16.4,<2.0.0',
 'datasets>=2.18.0,<3.0.0',
 'einops>=0.6.1,<0.7.0',
 'evaluate>=0.4.1,<0.5.0',
 'ftfy>=6.1.1,<7.0.0',
 'hvac>=1.1.1,<2.0.0',
 'iopath>=0.1.10,<0.2.0',
 'jupyter>=1.0.0,<2.0.0',
 'langchain-huggingface>=0.0.3,<0.0.4',
 'loguru>=0.7.0,<0.8.0',
 'numpy>=1.21,<2.0',
 'opencv-python>=4.10.0.84,<5.0.0.0',
 'peft==0.8.2',
 'platformdirs>=3.5.3,<4.0.0',
 'protobuf>=3.20.0,<4.0.0',
 'pydantic-settings>=2.2.1,<3.0.0',
 'pydantic>=2.7.0,<3.0.0',
 'python-dotenv>=1.0.0,<2.0.0',
 'pytorchvideo>=0.1.5,<0.2.0',
 'rouge-score>=0.1.2,<0.2.0',
 'scipy>=1.1.0,<2.0.0',
 'sentence-transformers>=2.2.2,<3.0.0',
 'sentencepiece>=0.1.99,<0.2.0',
 'soundfile>=0.12.1,<0.13.0',
 'timm>=0.9.7,<0.10.0',
 'tomlkit>=0.11.8,<0.12.0',
 'torch==2.3.1',
 'torchaudio>=2.0.2,<3.0.0',
 'transformers==4.43.1',
 'typer>=0.9.0,<0.10.0',
 'wandb>=0.15.3,<0.16.0',
 'wheel>=0.42.0,<0.43.0']

extras_require = \
{'deepspeed': ['deepspeed==0.12', 'accelerate==0.29.0'],
 'gpu': ['faiss-gpu>=1.7.2,<2.0.0', 'vllm==0.5.3']}

setup_kwargs = {
    'name': 'turbo-alignment',
    'version': '0.0.4',
    'description': 'turbo-alignment repository',
    'long_description': '# 🚀 Turbo-Alignment\n> Library for industrial alignment.\n\n\n## Table of Contents\n- [What is Turbo-Alignment?](#-what-is-turbo-alignment)\n- [Key Features](#-key-features)\n- [Supported Methods](#-supported-methods)\n- [Implemented metrics](#-implemented-metrics)\n- [How to Use](#-how-to-use)\n- [Installation](#-installation)\n- [Development](#-development)\n- [Library Roadmap](#-library-roadmap)\n- [FAQ](#-faq)\n- [License](#-license)\n\n<a name="-what-is-turbo-alignment"></a>\n## 🌟 What is Turbo-Alignment?\n\nTurbo-Alignment is a library designed to streamline the fine-tuning and alignment of large language models, leveraging advanced techniques to enhance efficiency and scalability.\n\n<a name="-key-features"></a>\n## ✨ Key Features\n\n- **📊 Comprehensive Metrics and Logging**: Includes a wide range of metrics such as self-bleu, KL divergence, diversity, etc. all supported out of the box.\n- **🛠️ Streamlined Method Deployment**: Simplifies the process of deploying new methods, allowing for quick development and integration of new datasets and trainers into your pipelines.\n- **📚 Ready-to-Use Examples**: Convenient examples with configurations and instructions for basic tasks.\n- **⚡ Fast Inference**: Optimized for quick inference using vLLM.\n- **🔄 End-to-End Pipelines**: From data preprocessing to model alignment.\n- **🌐 Multimodal Capabilities**: Extensive support for various multimodal functions like Vision Language Modeling.\n- **🔍 RAG Pipeline**: Unique pipeline for end2end retrieval-augmented generation training.\n\n<a name="-supported-methods"></a>\n## 🛠️ Supported Methods\n\n\nTurbo-Alignment supports a wide range of methods for model training and alignment, including:\n- **🎯** Supervised Fine-Tuning (SFT)\n- **🏆** Reward Modeling (RM)\n- **👍** Direct Preference Optimization (DPO)\n- **🧠** Kahneman & Tversky Optimization (KTO) Paired/Unpaired\n- **🔄** Contrastive Preference Optimization (CPO)\n- **🎭** Identity Preference Optimisation (IPO)\n- **🌟** Sequence Likelihood Calibration with Human Feedback (SLiC-HF)\n- **📊** Statistical Rejection Sampling Optimization (RSO)\n- **🌁** Vision Language Modeling using MLP from (LLaVA) or C-Abstractor from (HoneyBee) trainable projection model\n- **🗂️** Retrieval-Augmented Generation (RAG)\n\n<a name="-implemented-metrics"></a>\n## 🧮 Implemented Metrics\n- **🔠** Distinctness\n- **🌈** Diversity\n- **🔵** Self-BLEU\n- **➗** KL-divergence\n- **🏆** Reward\n- **📏** Length\n- **🌀** Perplexity\n- **🌟** METEOR\n- **🔍** Retrieval Utility\n\n<a name="-how-to-use"></a>\n## 🤖 How to Use\n\nTurbo-Alignment offers an intuitive interface for training and aligning large language models. Refer to the detailed examples and configuration files in the documentation to get started quickly with your specific use case. User-friendly guid available [here](docs/GUIDE.md).\n\nThe most crucial aspect is to prepare the dataset in the required format, after which the pipeline will handle everything automatically.\nExamples of datasets are available [here](docs/dataset_example.md).\n\n## Table of use-cases\n- [Training](#-train)\n  - [Supervised Fine-Tuning](#-sft-train)\n  - [Preference Tuning](#-preftune-train)\n    - [Reward Modeling](#-rm-train)\n    - [DPO, CPO, IPO, KTO (Paired)](#-dpo-train)\n    - [KTO (Unpaired)](#-kto-train)\n  - [Multimodal](#-multimodal-train)\n  - [RAG](#-rag-train)\n- [Inference](#-inference)\n  - [Supervised Fine-Tuning](#-sft-inference)\n  - [Multimodal](#-multimodal-inference)\n  - [RAG](#-rag-inference)\n- [Sampling](#-sampling)\n  - [Random](#-random-sampling)\n  - [RM](#-rm-sampling)\n  - [RSO](#-RSO-sampling)\n- [Common](#-common)\n  - [Preprocess](#-preprocess-common)\n  - [Merge adapters to base](#-merge-adapters-to-base-common)\n\n<a name="-train"></a>\n# Train\n\n<a name="-sft-train"></a>\n## Supervised Fine-Tuning\n- **📚 Dataset type** prepare your dataset  in the `ChatDataset`, examples available [here](docs/dataset_example.md#-chat-dataset) format.\n- **📝 Configs Example**: [sft.json](configs/exp/train/sft/sft.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment train_sft --experiment_settings_path configs/exp/train/sft/sft.json\n```\n<a name="-preftune-train"></a>\n## Preference Tuning\n<a name="-rm-train"></a>\n### Reward Modeling\n- **📚 Dataset type** prepare your dataset  in the `PairPreferencesDataset` format, examples available [here](docs/dataset_example.md#-pair-preferences)\n- **📝 Configs Example**: [rm.json](configs/exp/train/rm/rm.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment train_rm --experiment_settings_path configs/exp/train/rm/rm.json\n```\n\n<a name="-dpo-train"></a>\n### DPO, IPO, CPO, KTO (Paired)\n- **📚 Dataset type** prepare your dataset in the `PairPreferencesDataset` format, examples available [here](docs/dataset_example.md#pair-preferences)\n- **📝 Configs Example**: [dpo.json](configs/exp/train/dpo/dpo.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment train_dpo --experiment_settings_path configs/exp/train/dpo/dpo.json\n```\n\n<a name="-kto-train"></a>\n### KTO (Unpaired)\n- **📚 Dataset type** prepare your dataset in the `KTODataset` format, examples available [here](docs/dataset_example.md#-kto-dataset)\n- **📝 Configs Examples**: [kto.json](configs/exp/train/kto/kto.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment train_kto --experiment_settings_path configs/exp/train/kto/kto.json\n```\n\n<a name="-multimodal-train"></a>\n## Multimodal train\n⌛️  in progress..\n\n\n<a name="-rag-train"></a>\n## RAG (Retrieval-Augmented Generation) \n<a name="-sft-rag-train"></a>\n### SFT-RAG\n- **📚 Dataset type**: prepare your dataset in `ChatDataset`, examples available [here](docs/dataset_example.md#-chat-dataset) format.\n- **📝 Configs Example**: [sft_with_retrieval_utility](configs/exp/train/sft/llama/sft_with_retrieval_utility.json)\n- **🖥️ CLI launch command**: \n```bash\npython -m turbo_alignment train_sft --experiment_settings_path configs/exp/train/sft/llama/sft_with_retrieval_utility.json\n```\n<a name="-e2e-rag-train"></a>\n### End2End-RAG\n- **📚 Dataset type**: prepare your dataset in `ChatDataset`, examples available [here](docs/dataset_example.md#-chat-dataset) format.\n- **📝 Configs Example**: [end2end_rag](configs/exp/train/rag/end2end_rag.json)\n- **🖥️ CLI launch command**:\n```bash\npython -m turbo_alignment train_rag --experiment_settings_path configs/exp/train/rag/end2end_rag.json\n```\n\n<a name="-inference"></a>\n# Inference\n<a name="-chat-inference"></a>\n## Chat Inference\n- **📚 Dataset type** prepare your dataset  in the `ChatDataset`, examples available [here](docs/dataset_example.md#-chat-dataset) format.\n- **📝 Configs Example**: [sft.json](configs/exp/inference/generation/default_llama_adapter.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment inference_chat --inference_settings_path configs/exp/inference/generation/default_llama_adapter.json\n```\n\n<a name="-classification-inference"></a>\n## Classification Inference\n- **📚 Dataset type** prepare your dataset  in the `ClassificationDataset`, examples available [here](docs/dataset_example.md#-classification-dataset) format.\n- **📝 Configs Example**: [classification_inference.json](configs/exp/inference/classification/classification_inference.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment inference_classification --inference_settings_path configs/exp/train/sft/sft.json\n```\n\n<a name="-multimodal-inference"></a>\n## Multimodal Inference\n- **📚 Dataset type** prepare your dataset  in the `MultimodalDataset`, examples available [here](docs/dataset_example.md#-multimodal-dataset) format.\n- **📝 Configs Example**: [mlp.json](configs/exp/inference/multimodal/mlp.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment inference_multimodal --inference_settings_path configs/exp/inference/multimodal/mlp.json\n```\n\n<a name="-rag-inference"></a>\n## RAG Inference\n- **📚 Dataset type** prepare your dataset  in the `ChatDataset`, examples available [here](docs/dataset_example.md#-chat-dataset) format.\n- **📝 Configs Example**: [rag_inference.json](configs/exp/inference/rag/rag_inference.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment inference_rag --inference_settings_path configs/exp/inference/rag/rag_inference.json\n```\n\n<a name="-sampling"></a>\n# Sampling\n<a name="-random-sampling"></a>\n## Random Sampling\n- **📚 Dataset type** prepare your dataset  in the `SamplingRMDataset`, examples available [here](docs/dataset_example.md#-sampling-dataset) format.\n- **📝 Configs Example**: [random.json](tests/fixtures/configs/sampling/base.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment random_sample --experiment_settings_path tests/fixtures/configs/sampling/base.json\n```\n\n<a name="-rso-sampling"></a>\n## RSO Sampling\n- **📚 Dataset type** prepare your dataset  in the `SamplingRMDataset`, examples available [here](docs/dataset_example.md#-sampling-dataset) format.\n- **📝 Configs Example**: [rso.json](tests/fixtures/configs/sampling/rso.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment rso_sample --experiment_settings_path tests/fixtures/configs/sampling/rso.json\n```\n\n<a name="-rm-sampling"></a>\n## Reward Model Sampling\n- **📚 Dataset type** prepare your dataset  in the `SamplingRMDataset`, examples available [here](docs/dataset_example.md#-sampling-dataset) format.\n- **📝 Configs Example**: [rm.json](tests/fixtures/configs/sampling/rm.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment rm_sample --experiment_settings_path tests/fixtures/configs/sampling/rm.json\n```\n\n<a name="-common"></a>\n# Common\n<a name="-merge_adapters_to_base"></a>\n## Merge Adapters to base model\n- **📝 Configs Example**: [llama.json](configs/utils/merge_adapters_to_base/llama.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment merge_adapters_to_base --settings_path configs/utils/merge_adapters_to_base/llama.json\n```\n\n<a name="-preprocess_multimodal_dataset"></a>\n## Preprocess Multimodal Dataset\n- **📝 Configs Example**: [coco2014_clip.json](configs/utils/preprocess/coco2014_clip.json)\n- **🖥️ CLI launch command**\n```bash\npython -m turbo_alignment preprocess_multimodal_dataset --settings_path configs/utils/preprocess/coco2014_clip.json\n```\n\n\n<a name="-installation"></a>\n## 🚀 Installation\n\n### 📦 Python Package\n```bash\npip install turbo-alignment\n```\n\n### 🛠️ From Source\nFor the latest features before an official release:\n```bash\npip install git+https://github.com/turbo-llm/turbo-alignment.git\n```\n\n### 📂 Repository\nClone the repository for access to examples:\n```bash\ngit clone https://github.com/turbo-llm/turbo-alignment.git\n```\n\n<a name="-development"></a>\n## 🌱 Development\n\nContributions are welcome! Read the [contribution guide](https://github.com/turbo-llm/turbo-alignment/blob/main/CONTRIBUTING.md) and set up the development environment:\n```bash\ngit clone https://github.com/turbo-llm/turbo-alignment.git\ncd turbo-alignment\npoetry install\n```\n\n<a name="-library-roadmap"></a>\n## 📍 Library Roadmap\n\n- Increasing number of tutorials\n- Enhancing test coverage\n- Implementation of Online RL methods like PPO and Reinforce\n- Facilitating distributed training\n- Incorporating low-memory training approaches\n\n\n## ❓ FAQ\n### How do I install Turbo-Alignment?\nSee the [Installation](#-installation) section for detailed instructions.\n\n### Where can I find docs?\nGuides and docs are available [here](docs/GUIDE.md).\n\n### Where can I find tutorials?\nTutorials are available [here](tutorials/tutorial.md).\n\n\n## 📝 License\nThis project is licensed, see the [LICENSE](https://github.com/turbo-llm/turbo-alignment/-/blob/main/LICENSE) file for details.\n\n\n## References\n\n- DPO Trainer implementation inspired by Leandro von Werra et al. (2020) TRL: Transformer Reinforcement Learning. GitHub repository, GitHub. Available at: [https://github.com/huggingface/trl](https://github.com/huggingface/trl).\n\n- Registry implementation inspired by Matt Gardner, Joel Grus, Mark Neumann, Oyvind Tafjord, Pradeep Dasigi, Nelson F. Liu, Matthew Peters, Michael Schmitz, and Luke S. Zettlemoyer. 2017. AllenNLP: A Deep Semantic Natural Language Processing Platform. Available at: [arXiv:1803.07640](https://arxiv.org/abs/1803.07640).\n\n- Liger Kernels implementation inspired by Hsu, Pin-Lun, Dai, Yun, Kothapalli, Vignesh, Song, Qingquan, Tang, Shao, and Zhu, Siyu, 2024. Liger-Kernel: Efficient Triton Kernels for LLM Training. Available at: [https://github.com/linkedin/Liger-Kernel](https://github.com/linkedin/Liger-Kernel).\n',
    'author': 'T Mega Alignment Team',
    'author_email': 'n.surnachev@tcsbank.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/turbo-llm/turbo-alignment',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from typing import Iterable

from accelerate import Accelerator
from transformers import PreTrainedModel, PreTrainedTokenizerBase

from turbo_alignment.cherry_picks.base import CherryPickCallbackBase
from turbo_alignment.dataset.chat import InferenceChatDataset
from turbo_alignment.generators.chat import ChatGenerator
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import KLType
from turbo_alignment.metrics.utils import get_logits
from turbo_alignment.settings.cherry_pick import ChatCherryPickSettings
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults


class ChatCherryPickCallback(CherryPickCallbackBase[InferenceChatDataset]):
    def __init__(
        self,
        cherry_pick_settings: ChatCherryPickSettings,
        datasets: Iterable[InferenceChatDataset],
        metrics: list[Metric],
    ) -> None:
        super().__init__(cherry_pick_settings=cherry_pick_settings, datasets=datasets, metrics=metrics)
        self._custom_generation_settings = cherry_pick_settings.custom_generation_settings
        self._generator_transformers_settings = cherry_pick_settings.generator_transformers_settings

    def _get_dataset_metrics(
        self,
        dataset: InferenceChatDataset,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> list[MetricResults]:
        accelerator: Accelerator = kwargs.get('accelerator', None)
        metrics_kwargs: dict = kwargs.get('metrics_kwargs', {})
        ref_model: dict = kwargs.get('ref_model', None)
        sft_model: dict = kwargs.get('sft_model', None)

        generator = ChatGenerator(
            model=model,
            tokenizer=tokenizer,
            transformers_settings=self._generator_transformers_settings,
            custom_generation_settings=self._custom_generation_settings,
            accelerator=accelerator,
            return_logits=True,
        )

        batch_size = self._generator_transformers_settings.num_return_sequences

        generations = generator.generate_from_dataset(dataset)

        prompts = [record['prompt'] for record in dataset]
        string_answers = [[answer.content for answer in g.answers] for g in generations]
        string_labels = [[g.messages[-1].content] * len(g.answers) for g in generations]

        flattened_answers = [answer for g in generations for answer in g.answers]

        answer_tokens_ids = [answer.answer_token_ids.cpu() for answer in flattened_answers]
        input_tokens_ids = [answer.input_token_ids.cpu() for answer in flattened_answers]

        logits = [answer.logits[:, answer.input_token_ids.size(-1) :, :].cpu() for answer in flattened_answers]

        if ref_model is not None:
            metrics_kwargs[KLType.REFERENCE_MODEL] = get_logits(input_tokens_ids, answer_tokens_ids, ref_model)

        if sft_model is not None:
            metrics_kwargs[KLType.SFT_MODEL] = get_logits(input_tokens_ids, answer_tokens_ids, sft_model)

        metric_outputs = [
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(
                        label=dataset.source.name + '@@' + 'prompt',
                        values=[prompt for prompt in prompts for _ in range(batch_size)],
                    )
                ]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'labels', values=string_labels)
                ]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'answer', values=string_answers)
                ]
            ),
        ]

        for metric in self._metrics:
            metric_results = metric.compute(
                model=model,
                dataset=dataset,
                references=string_labels,
                predictions=string_answers,
                accelerator=accelerator,
                logits=logits,
                answer_tokens_ids=answer_tokens_ids,
                tokenizer=tokenizer,
                input_token_ids=input_tokens_ids,
                dataset_name=dataset.source.name,
                metrics_kwargs=metrics_kwargs,
            )

            metric_outputs.extend(metric_results)
        return metric_outputs



import os
import random

import numpy as np
import torch


def set_random_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)
    os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:2'
    os.environ['PL_GLOBAL_SEED'] = str(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True



from pathlib import Path

import torch
from safetensors import safe_open

from turbo_alignment.common.data.multimodal.image.base import BaseImageReader
from turbo_alignment.common.data.multimodal.registry import (
    AudioModalityReaderRegistry,
    ImageModalityReaderRegistry,
)
from turbo_alignment.settings.modality import ModalityReader


@AudioModalityReaderRegistry.register(ModalityReader.PICKLE)
@ImageModalityReaderRegistry.register(ModalityReader.PICKLE)
class FileReader(BaseImageReader):
    def __init__(self, **_kwargs):
        self.processed_tensors = None

    @staticmethod
    def _get_safetensors_file(path: Path) -> Path:
        return list(path.glob('*.safetensors'))[0]  # FIXME: What if there is more than one safetensors file?

    def read(self, path: str) -> torch.Tensor:
        safetensors_file = self._get_safetensors_file(Path(path).parent)
        if self.processed_tensors is None:
            self.processed_tensors = safe_open(safetensors_file, framework='pt', device='cpu')
        return self.processed_tensors.get_tensor(Path(path).name)



from turbo_alignment.settings.cherry_pick import ChatCherryPickSettings
from turbo_alignment.settings.datasets.chat import ChatMultiDatasetSettings
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings
from turbo_alignment.settings.rag_model import RAGPreTrainedModelSettings


class RAGTrainExperimentSettings(BaseTrainExperimentSettings):
    model_settings: RAGPreTrainedModelSettings  # type: ignore[assignment]
    train_dataset_settings: ChatMultiDatasetSettings
    val_dataset_settings: ChatMultiDatasetSettings

    cherry_pick_settings: ChatCherryPickSettings



from pathlib import Path
from typing import Optional

import torch
from transformers import CLIPModel

from turbo_alignment.modeling.multimodal.encoders.image.base import BaseImageEncoder
from turbo_alignment.modeling.multimodal.encoders.registry import (
    ModalityEncoderRegistry,
)
from turbo_alignment.settings.modality import ModalityEncoderType


@ModalityEncoderRegistry.register(ModalityEncoderType.CLIP)
class CLIPImageModeling(BaseImageEncoder):
    def __init__(self, encoder_path: Path, model_clip: Optional[CLIPModel] = None, is_pickle: bool = False):
        super().__init__()
        if model_clip is not None:
            self.model_clip = model_clip
        else:
            self.model_clip = CLIPModel.from_pretrained(encoder_path)
        self.is_pickle = is_pickle

    @staticmethod
    def _get_clip_hidden_states(model_clip: CLIPModel, inputs: torch.Tensor, is_pickle: bool = False) -> torch.Tensor:
        if is_pickle:
            return inputs
        # https://github.com/huggingface/transformers/blob/main/src/transformers/models/llava/modeling_llava.py#L213
        # -2 is default value of vision_feature_layer in llava config
        # [1:] is everything after vit [cls] token
        return model_clip.vision_model(inputs.squeeze(1), output_hidden_states=True).hidden_states[-2][
            :, 1:
        ]  # FIXME: squeeze dimension?

    def encode(self, inputs: torch.Tensor) -> torch.Tensor:
        return self._get_clip_hidden_states(self.model_clip, inputs, self.is_pickle)

    @property
    def emb_dim(self):
        return self.model_clip.config.vision_config.hidden_size

    @property
    def device(self):
        return self.model_clip.device

    @property
    def n_modality_embs(self) -> int:
        image_size = self.model_clip.config.vision_config.image_size
        dummy_pixel_values = torch.empty(1, 3, image_size, image_size)
        hidden_states = self._get_clip_hidden_states(self.model_clip, dummy_pixel_values, is_pickle=False)
        return hidden_states.shape[1]



from turbo_alignment.settings.cherry_pick import ChatCherryPickSettings
from turbo_alignment.settings.datasets.ddpo import DDPOMultiDatasetSettings
from turbo_alignment.settings.model import (
    PreTrainedAdaptersModelSettings,
    PreTrainedModelSettings,
)
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings
from turbo_alignment.settings.tf.tokenizer import TokenizerSettings


class DDPOTrainExperimentSettings(BaseTrainExperimentSettings):
    beta: float = 0.1
    forward_kl: bool = False

    use_ref_model: bool = True
    rm_settings: PreTrainedModelSettings | PreTrainedAdaptersModelSettings

    train_dataset_settings: DDPOMultiDatasetSettings
    val_dataset_settings: DDPOMultiDatasetSettings

    cherry_pick_settings: ChatCherryPickSettings

    rm_tokenizer_settings: TokenizerSettings



from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Callable, Literal

import torch
from torch import nn
from torch.utils.data import Dataset
from transformers import (
    DefaultFlowCallback,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    PrinterCallback,
    ProgressCallback,
    TrainerCallback,
    TrainerControl,
    TrainingArguments,
)
from transformers.integrations import get_reporting_integration_callbacks

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.tf.callbacks.common import MetricsCallbackHandler
from turbo_alignment.trainers.dpo import DPOTrainer
from turbo_alignment.trainers.utils import concatenated_inputs, prepare_model

logger = get_project_logger()


@dataclass
class DDPOTrainingArguments(TrainingArguments):
    use_ref_model: bool = True
    average_log_prob: bool = False

    beta: float = 0.1
    forward_kl: bool = False


class DDPOTrainer(DPOTrainer):
    """
    Inspired by NLP Research

    """

    def __init__(
        self,
        model: PreTrainedModel | nn.Module,
        data_collator: Callable,
        args: DDPOTrainingArguments,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        rm: PreTrainedModel | nn.Module,
        ref_model: PreTrainedModel | nn.Module | None = None,
        tokenizer: PreTrainedTokenizerBase | None = None,
        callbacks: list[TrainerCallback] | None = None,
        **kwargs,
    ):
        self.data_collator = data_collator

        self.average_log_prob = args.average_log_prob

        self._stored_metrics: dict[str, dict[str, list[float]]] = defaultdict(lambda: defaultdict(list))
        self.beta = args.beta
        self.forward_kl = args.forward_kl

        super().__init__(
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            callbacks=callbacks,
            **kwargs,
        )

        default_callbacks = [DefaultFlowCallback] + get_reporting_integration_callbacks(self.args.report_to)
        callbacks = default_callbacks if callbacks is None else default_callbacks + callbacks
        self.callback_handler = MetricsCallbackHandler(
            callbacks, model, tokenizer, None, None, ref_model=ref_model, accelerator=self.accelerator
        )
        self.add_callback(PrinterCallback if self.args.disable_tqdm else ProgressCallback)
        self.control: TrainerControl = self.callback_handler.on_init_end(self.args, self.state, self.control)

        self.ref_model = ref_model
        self.rm = rm

        if self.ref_model is not None:
            self.ref_model = prepare_model(self.ref_model, self.accelerator, self.is_deepspeed_enabled)

        self.rm = prepare_model(self.rm, self.accelerator, self.is_deepspeed_enabled)

    def ddpo_loss(
        self,
        policy_chosen_logps: torch.Tensor,
        policy_rejected_logps: torch.Tensor,
        reference_chosen_logps: torch.Tensor,
        reference_rejected_logps: torch.Tensor,
        rm_reward_chosen: torch.Tensor,
        rm_reward_rejected: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        pi_logratios = policy_chosen_logps - policy_rejected_logps
        ref_logratios = reference_chosen_logps - reference_rejected_logps

        logits = pi_logratios - ref_logratios

        chosen_rewards = (self.beta * (policy_chosen_logps - reference_chosen_logps)).detach()
        rejected_rewards = (self.beta * (policy_rejected_logps - reference_rejected_logps)).detach()
        p_sft = torch.sigmoid(self.beta * logits)

        p_rm = torch.exp(rm_reward_chosen) / (torch.exp(rm_reward_chosen) + torch.exp(rm_reward_rejected))

        p1, p2 = p_sft, p_rm
        if self.forward_kl:
            p1, p2 = p_rm, p_sft

        loss = (1 - p1) * (torch.log(1 - p1) - torch.log(1 - p2)) + p1 * (torch.log(p1) - torch.log(p2))
        return loss, chosen_rewards, rejected_rewards

    def concatenated_sft_forward(
        self, model: nn.Module, batch: dict[str, Any]
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        concatenated_batch = concatenated_inputs(batch, prefix='sft_', device=self.accelerator.device)
        all_logits = model(
            concatenated_batch['input_ids'],
            attention_mask=concatenated_batch['attention_mask'],
        ).logits.to(torch.float32)
        all_logps = self._get_batch_logps(
            all_logits,
            concatenated_batch['labels'],
            average_log_prob=self.average_log_prob,
        )
        chosen_idxs = batch['sft_inputs_w']['input_ids'].shape[0]

        chosen_logps = all_logps[:chosen_idxs]
        rejected_logps = all_logps[chosen_idxs:]

        chosen_logits = all_logits[:chosen_idxs]
        rejected_logits = all_logits[chosen_idxs:]
        return chosen_logps, rejected_logps, chosen_logits, rejected_logits

    def concatenated_rm_forward(self, batch: dict[str, Any]) -> tuple[torch.Tensor, torch.Tensor]:
        concatenated_batch = concatenated_inputs(batch, prefix='rm_', device=self.accelerator.device)
        with torch.no_grad():
            rewards = self.rm(
                input_ids=concatenated_batch['input_ids'], attention_mask=concatenated_batch['attention_mask']
            ).logits

        chosen_idxs = batch['rm_inputs_w']['input_ids'].shape[0]

        chosen_rewards, rejected_rewards = rewards[:chosen_idxs, :], rewards[chosen_idxs:, :]
        return chosen_rewards, rejected_rewards

    def _get_ref_logps(self, batch: dict[str, Any]) -> tuple[torch.Tensor, torch.Tensor]:
        with torch.no_grad():
            if self.ref_model is not None:
                (
                    reference_chosen_logps,
                    reference_rejected_logps,
                    _,
                    _,
                ) = self.concatenated_sft_forward(self.ref_model, batch)
            else:
                with self.accelerator.unwrap_model(self.model).disable_adapter():
                    (
                        reference_chosen_logps,
                        reference_rejected_logps,
                        _,
                        _,
                    ) = self.concatenated_sft_forward(self.model, batch)
        return reference_chosen_logps, reference_rejected_logps

    def get_batch_metrics(
        self,
        model: nn.Module,
        batch: dict[str, Any],
        train_eval: Literal['train', 'eval'] = 'train',
    ) -> tuple[torch.Tensor, dict[str, float]]:
        metrics = {}

        (
            policy_chosen_logps,
            policy_rejected_logps,
            policy_chosen_logits,
            policy_rejected_logits,
        ) = self.concatenated_sft_forward(model, batch)

        reference_chosen_logps, reference_rejected_logps, *_ = self._get_ref_logps(batch)
        rm_rewards_chosen, rm_rewards_rejected = self.concatenated_rm_forward(batch)

        losses, chosen_rewards, rejected_rewards = self.ddpo_loss(
            policy_chosen_logps,
            policy_rejected_logps,
            reference_chosen_logps,
            reference_rejected_logps,
            rm_rewards_chosen,
            rm_rewards_rejected,
        )

        sft_reward_accuracies = (chosen_rewards > rejected_rewards).float()
        rm_reward_accuracies = (rm_rewards_chosen > rm_rewards_rejected).float()

        prefix = 'eval_' if train_eval == 'eval' else ''
        metrics[f'{prefix}sft_rewards/chosen'] = chosen_rewards.detach().cpu().mean().item()
        metrics[f'{prefix}sft_rewards/rejected'] = rejected_rewards.detach().cpu().mean().item()
        metrics[f'{prefix}sft_rewards/accuracies'] = sft_reward_accuracies.detach().cpu().mean().item()
        metrics[f'{prefix}sft_rewards/margins'] = (chosen_rewards - rejected_rewards).detach().cpu().mean().item()

        metrics[f'{prefix}rm_rewards/chosen'] = rm_rewards_chosen.detach().cpu().mean().item()
        metrics[f'{prefix}rm_rewards/rejected'] = rm_rewards_rejected.detach().cpu().mean().item()
        metrics[f'{prefix}rm_rewards/accuracies'] = rm_reward_accuracies.detach().cpu().mean().item()
        metrics[f'{prefix}rm_rewards/margins'] = (rm_rewards_chosen - rm_rewards_rejected).detach().cpu().mean().item()

        metrics[f'{prefix}logps/rejected'] = policy_rejected_logps.detach().cpu().mean().item()
        metrics[f'{prefix}logps/chosen'] = policy_chosen_logps.detach().cpu().mean().item()
        metrics[f'{prefix}logits/rejected'] = policy_rejected_logits.detach().cpu().mean().item()
        metrics[f'{prefix}logits/chosen'] = policy_chosen_logits.detach().cpu().mean().item()

        return losses.nanmean(), metrics



from typing import Generator

import torch
from accelerate import Accelerator
from transformers import PreTrainedModel, PreTrainedTokenizerBase

from turbo_alignment.common.tf.loaders.model.model import load_model
from turbo_alignment.generators.base import BaseGenerator
from turbo_alignment.generators.rag import RagGenerator
from turbo_alignment.modeling.rag.rag_model import RagSequenceForGeneration
from turbo_alignment.modeling.rag.rag_tokenizer import RagTokenizer
from turbo_alignment.pipelines.inference.base import BaseInferenceStrategy
from turbo_alignment.settings.pipelines.inference import RAGInferenceExperimentSettings
from turbo_alignment.settings.rag_model import RAGPreTrainedModelSettings


class RAGInferenceStrategy(BaseInferenceStrategy):
    @staticmethod
    def _load_model(
        model_settings: RAGPreTrainedModelSettings,
        tokenizer: PreTrainedTokenizerBase,
    ) -> torch.nn.Module | PreTrainedModel:
        generator = load_model(model_settings.generator_settings, tokenizer.generator)
        question_encoder = load_model(model_settings.question_encoder_settings, tokenizer.question_encoder)

        model = RagSequenceForGeneration(model_settings, generator, question_encoder, tokenizer)
        return model

    def _get_single_inference_settings(
        self, experiment_settings: RAGInferenceExperimentSettings, accelerator: Accelerator
    ) -> Generator[tuple[PreTrainedTokenizerBase, BaseGenerator, str, dict], None, None]:
        save_file_id = 0

        for model_inference_settings in experiment_settings.inference_settings:
            tokenizer = RagTokenizer(
                model_settings=model_inference_settings.model_settings,
                tokenizer_path=model_inference_settings.tokenizer_settings.tokenizer_path,
                use_fast=model_inference_settings.tokenizer_settings.use_fast,
            )

            model = self._load_model(model_inference_settings.model_settings, tokenizer).to(accelerator.device)
            model.eval()

            for generation_settings in model_inference_settings.generation_settings:
                generator = RagGenerator(
                    transformers_settings=generation_settings.transformers_settings,
                    custom_generation_settings=generation_settings.custom_settings,
                    tokenizer=tokenizer,
                    model=model,
                    accelerator=accelerator,
                )

                parameters_to_save = {
                    'model_settings': model_inference_settings.model_settings.dict(),
                    'generation_settings': generation_settings.dict(),
                }

                yield tokenizer, generator, f'single_inference_{save_file_id}.jsonl', parameters_to_save



# pylint: disable=unused-import
import torch

from turbo_alignment.common.singleton import ParametrizedSingleton
from turbo_alignment.modeling.imagebind.heads.registry import Heads
from turbo_alignment.modeling.imagebind.models import (
    ImageBindArchitectureSettings,
    ImageBindSettings,
)
from turbo_alignment.modeling.imagebind.postprocessors.registry import Postprocessors
from turbo_alignment.modeling.imagebind.preprocessors.registry import Preprocessors
from turbo_alignment.modeling.imagebind.trunks.registry import Trunks


class ImageBindModel(torch.nn.Module):
    def __init__(self, settings: ImageBindArchitectureSettings):
        super().__init__()

        self.modality_heads = Heads(settings)
        self.modality_trunks = Trunks(settings)
        self.modality_preprocessors = Preprocessors(settings)
        self.modality_postprocessors = Postprocessors(settings)

    def forward(self, inputs):
        outputs = {}
        for modality_key, modality_value in inputs.items():
            reduce_list = modality_value.ndim >= 5  # Audio and Video inputs consist of multiple clips
            if reduce_list:
                B, S = modality_value.shape[:2]
                modality_value = modality_value.reshape(B * S, *modality_value.shape[2:])

            if modality_value is not None:
                modality_value = self.modality_preprocessors[modality_key](**{modality_key: modality_value})
                trunk_inputs = modality_value['trunk']
                head_inputs = modality_value['head']
                modality_value = self.modality_trunks[modality_key](**trunk_inputs)
                modality_value = self.modality_heads[modality_key](modality_value, **head_inputs)
                modality_value = self.modality_postprocessors[modality_key](modality_value)

                if reduce_list:
                    modality_value = modality_value.reshape(B, S, -1)
                    modality_value = modality_value.mean(dim=1)

                outputs[modality_key] = modality_value

        return outputs


def load_imagebind(settings: ImageBindSettings) -> ImageBindModel:
    model = ImageBindModel(settings.architecture_settings)

    if settings.weights_path:
        model.load_state_dict(torch.load(settings.weights_path), strict=False)

    if not settings.is_trainable:
        for param in model.parameters():
            param.requires_grad = False

    return model


class ImageBindSingleton(metaclass=ParametrizedSingleton[ImageBindSettings]):  # type: ignore[misc]
    def __init__(self, settings: ImageBindSettings) -> None:
        self._model = load_imagebind(settings)

    def get(self) -> ImageBindModel:
        return self._model



import random
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Generic, TypeVar, overload

import torch
from torch.utils.data import Dataset
from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.dataset.base.models import DatasetRecord
from turbo_alignment.settings.datasets.base import (
    BaseDatasetSettings,
    DatasetSourceSettings,
)

logger = get_project_logger()


RecordT = TypeVar('RecordT', bound=DatasetRecord)


class BaseDataset(Dataset, ABC, Generic[RecordT]):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: BaseDatasetSettings,
    ) -> None:
        self.source = source
        self.settings = settings

        self.original_records_map: dict[str, RecordT] = {}
        self.records: list[dict[str, torch.Tensor]] = []

    def _read(self) -> None:
        if self.source.records_data:
            records = self._read_records(self.source.records_data)
        elif self.source.records_path:
            records = self._read_records(self.source.records_path)
        else:
            raise ValueError('At least one of records_data and records_path should be not None')

        if self.source.offset is not None and self.source.n_rows is not None:
            records = records[self.source.offset : self.source.offset + self.source.n_rows]

        self.original_records_map, self.records = self._sample_dataset(records)

        logger.info(f'Sampled {len(self.records)} records with offset {self.source.offset}')

    def _sample_dataset(
        self,
        original_records: list[RecordT],
    ) -> tuple[dict[str, RecordT], list[dict[str, Any]]]:
        if self.source.sample_rate is not None:
            logger.info(f'Sampling dataset {self.source.name} with sample rate: {self.source.sample_rate}')
            sampled_original_records = {
                record.id: record for record in original_records if random.random() <= self.source.sample_rate
            }
        elif self.source.num_samples is not None:
            logger.info(f'Sampling {self.source.num_samples} from dataset {self.source.name}')
            sampled_original_records = {
                record.id: record
                for record in random.sample(original_records, k=min(self.source.num_samples, len(original_records)))
            }
        else:
            raise ValueError('neither sample_rate nor num_samples are not set')

        sampled_records = [r for r in self.convert_records(list(sampled_original_records.values())) if r is not None]

        return sampled_original_records, sampled_records

    def __len__(self) -> int:
        return len(self.records)

    def __getitem__(self, index: int) -> dict[str, Any]:
        return self.records[index]

    def __iter__(self):
        return iter(self.records)

    def get_original_record_by_id(self, record_id: str) -> RecordT:
        return self.original_records_map[record_id]

    @abstractmethod
    def convert_records(self, records: list[RecordT]) -> list[dict[str, Any] | None]:
        ...

    @staticmethod
    @abstractmethod
    @overload
    def _read_records(records: Path) -> list[RecordT]:
        ...

    @staticmethod
    @abstractmethod
    @overload
    def _read_records(records: list[dict]) -> list[RecordT]:
        ...

    @staticmethod
    @abstractmethod
    def _read_records(records):
        ...


class AlignmentDataset(BaseDataset, ABC, Generic[RecordT]):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: BaseDatasetSettings,
        tokenizer: PreTrainedTokenizerBase,
    ) -> None:
        super().__init__(source=source, settings=settings)

        self.tokenizer = tokenizer
        self._logged = False

    def _log_example(self, prompt: str, answer: str | None = None) -> None:
        if not self._logged:
            message = f'Source and target examples:\n' f'Prompt: {prompt}\n'
            if answer:
                message += f'Answer: {answer}'
            logger.info(message)
            self._logged = True



from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.dataset.chat.models import ChatMessage, ChatMessageRole

logger = get_project_logger()


class Conversation:
    def __init__(
        self,
        system_prompt: str | None,
        messages: list[ChatMessage],
        ignore_system_prompt: bool,
    ):
        self._messages: list[ChatMessage] = []

        if system_prompt and ignore_system_prompt:
            raise ValueError('You can not ignore system_prompt and use system prompt in config simultaneously')

        if system_prompt or ignore_system_prompt:
            if messages[0].role == ChatMessageRole.SYSTEM:
                messages = messages[1:]

            if system_prompt:
                self._messages += [ChatMessage(role=ChatMessageRole.SYSTEM, content=system_prompt, disable_loss=True)]

        self._messages += messages

    @property
    def messages(self) -> list[ChatMessage]:
        return self._messages

    def get_prompt_repr(self, left_bound: int, right_bound: int) -> str:
        prompt: str = ''
        if self._messages[0].role == ChatMessageRole.SYSTEM and left_bound != 0:
            prompt += f'[{self._messages[0].role}: {self._messages[0].content}]'

        for msg in self._messages[left_bound:right_bound]:
            prompt += f'[{msg.role}: {msg.content}]'
        return prompt

    def __len__(self) -> int:
        return len(self._messages)



from typing import Literal

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.datasets.base import (
    BaseDatasetSettings,
    DatasetType,
    MultiDatasetSettings,
)


class ChatPromptTemplate(ExtraFieldsNotAllowedBaseModel):
    prefix_template: str
    suffix_template: str
    role_tag_mapping: dict[str, str]


class ChatDatasetSettings(BaseDatasetSettings):
    dataset_type: Literal[DatasetType.CHAT] = DatasetType.CHAT

    # FIXME: this settings are strange for inference dataset
    only_last_replica_loss: bool = False
    only_answer_loss: bool = True
    random_cut: bool = False  # TODO: not common property for train/inference

    keep_end: bool | None = None
    max_tokens_count: int | None
    prompt_template: ChatPromptTemplate
    ignore_system_prompt: bool = False


class ChatMultiDatasetSettings(ChatDatasetSettings, MultiDatasetSettings):
    ...



from typing import Any

import torch
from torch import nn
from transformers import DataCollatorWithPadding, PreTrainedTokenizerBase

from turbo_alignment.dataset.pair_preferences import PairPreferenceRecord
from turbo_alignment.dataset.sampling.models import SamplingDatasetRecord
from turbo_alignment.generators.base import BaseGenerator
from turbo_alignment.settings.generators.outputs.rm import (
    RMPairInferenceOutput,
    RMSamplingInferenceOutput,
)


class RMPairGenerator(BaseGenerator[PairPreferenceRecord, RMPairInferenceOutput]):
    def __init__(self, tokenizer: PreTrainedTokenizerBase, **kwargs):
        self._collator = DataCollatorWithPadding(tokenizer=tokenizer)

        super().__init__(tokenizer=tokenizer, **kwargs)

    def _generate_from_batch(
        self, records: list[dict[str, Any]], original_records: list[PairPreferenceRecord], dataset_name: str
    ) -> list[RMPairInferenceOutput]:
        merged_inputs = [r['inputs_w'] for r in records] + [r['inputs_l'] for r in records]
        batch = self._collator(merged_inputs)
        input_ids = batch['input_ids'].to(self.device)
        attn_mask = batch['attention_mask'].to(self.device)

        with torch.no_grad():
            rewards = self._model(input_ids=input_ids, attention_mask=attn_mask).logits.cpu()
            rewards_w, rewards_l = rewards[: len(records)], rewards[len(records) :]

        return [
            RMPairInferenceOutput(
                id=record.id,
                context=record.context,
                answer_w=record.answer_w,
                answer_l=record.answer_l,
                reward_w=reward_w.item(),
                reward_l=reward_l.item(),
                dataset_name=dataset_name,
            )
            for record, reward_w, reward_l in zip(original_records, rewards_w, rewards_l)
        ]


class RMSamplingGenerator(BaseGenerator[SamplingDatasetRecord, RMSamplingInferenceOutput]):
    def __init__(self, tokenizer: PreTrainedTokenizerBase, micro_batch: int, **kwargs):
        self._collator = DataCollatorWithPadding(tokenizer=tokenizer)
        self._micro_batch = micro_batch
        super().__init__(tokenizer=tokenizer, **kwargs)

    def _generate_from_batch(
        self, records: list[dict[str, Any]], original_records: list[SamplingDatasetRecord], dataset_name: str
    ) -> list[RMSamplingInferenceOutput]:
        merged_inputs = [inputs for record in records for key, inputs in record['answers'].items()]

        if len(merged_inputs) == 0:
            return []

        rewards = []
        with torch.no_grad():
            input_ids = nn.utils.rnn.pad_sequence(
                [item['input_ids'] for item in merged_inputs],
                padding_value=self._tokenizer.pad_token_id,
                batch_first=True,
            )
            attn_mask = nn.utils.rnn.pad_sequence(
                [item['attention_mask'] for item in merged_inputs],
                padding_value=0,
                batch_first=True,
            )
            for i in range(0, len(input_ids), self._micro_batch):
                input_ids_batch = input_ids[i : i + self._micro_batch].to(self.device)
                attn_mask_batch = attn_mask[i : i + self._micro_batch].to(self.device)
                rewards.extend(self._model(input_ids=input_ids_batch, attention_mask=attn_mask_batch).logits.cpu())

        rewards = torch.cat(rewards, dim=0)

        record_rewards = [
            {answer_id: reward.item() for answer_id, reward in zip(record['answers'].keys(), rewards)}
            for record in records
        ]

        return [
            RMSamplingInferenceOutput(
                id=record.id,
                rewards=rewards,
                messages=record.messages,
                dataset_name=dataset_name,
                answers=record.answers,
            )
            for record, rewards in zip(original_records, record_rewards)
        ]



import abc

from turbo_alignment.common.registry import Registrable
from turbo_alignment.metrics.registry import MetricSettings
from turbo_alignment.settings.metric import MetricResults


class Metric(abc.ABC, Registrable):
    def __init__(self, settings: MetricSettings) -> None:
        self._settings: MetricSettings = settings

    @abc.abstractmethod
    def compute(self, **kwargs) -> list[MetricResults]:
        ...



from typing import Sequence

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.generators.chat import CustomChatGenerationSettings
from turbo_alignment.settings.pipelines.inference.base import (
    InferenceExperimentSettings,
    SingleModelInferenceSettings,
)
from turbo_alignment.settings.tf.generation import GeneratorTransformersSettings


class ChatGenerationSettings(ExtraFieldsNotAllowedBaseModel):
    transformers_settings: GeneratorTransformersSettings
    custom_settings: CustomChatGenerationSettings


class ChatSingleModelInferenceSettings(SingleModelInferenceSettings):
    generation_settings: list[ChatGenerationSettings]
    use_vllm: bool = False
    tensor_parallel_size: int = 1


class ChatInferenceExperimentSettings(InferenceExperimentSettings):
    inference_settings: Sequence[ChatSingleModelInferenceSettings]



import torch
from transformers import PreTrainedModel


def calculate_cross_entropy(logits: torch.Tensor, labels: torch.Tensor, pad_token_id: int, reduction: str):
    shift_logits = logits[:, :-1, :].contiguous().to(torch.float32)
    shift_labels = labels[:, 1:].contiguous().long()
    mask = (shift_labels != pad_token_id).to(torch.int)
    valid_length = mask.sum(dim=-1)

    ce_func = torch.nn.CrossEntropyLoss(ignore_index=pad_token_id, reduction=reduction)
    ce = ce_func(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))

    ce = ce.view(shift_labels.size(0), -1)
    ce = torch.sum(ce * mask, -1) / valid_length
    return ce


def logprobs_from_logits(logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
    logp = torch.nn.functional.log_softmax(logits, dim=-1)
    logps = torch.gather(logp, dim=-1, index=labels.unsqueeze(-1)).squeeze(-1)
    return logps


def get_logits(
    input_tokens_ids: list[torch.Tensor], answer_tokens_ids: list[torch.Tensor], model: PreTrainedModel
) -> list[torch.Tensor]:
    logits = []

    for input_token_ids, answer_token_ids in zip(input_tokens_ids, answer_tokens_ids):
        with torch.no_grad():
            input_ = torch.cat((input_token_ids, answer_token_ids), dim=-1)  # calculate logits on input + answer
            logit = model(input_.to(model.device)).logits
            logit = logit[:, input_token_ids.size(-1) :, :].cpu()
            logits.append(logit)  # return only answer logits
    return logits



from accelerate.utils import is_deepspeed_available
from torch import nn
from transformers import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)

from turbo_alignment.settings.pipelines.train.dpo import SyncRefModelSettings


class SyncRefModelCallback(TrainerCallback):
    def __init__(self, *args, sync_ref_settings: SyncRefModelSettings, **kwargs) -> None:
        self._sync_ref_settings = sync_ref_settings

        super().__init__(*args, **kwargs)

    @staticmethod
    def _sync_target_model(model: nn.Module, target_model: nn.Module, alpha: float) -> None:
        for target_param, copy_param in zip(target_model.parameters(), model.parameters()):
            target_param.data.copy_((alpha * copy_param.data) + (1.0 - alpha) * target_param.data)

    @classmethod
    def sync_target_model(
        cls, model: nn.Module, target_model: nn.Module, alpha: float, is_zero3_enabled: bool = False
    ) -> None:
        if is_deepspeed_available() and is_zero3_enabled:
            import deepspeed

            with deepspeed.zero.GatheredParameters(list(model.parameters()), modifier_rank=0):
                if deepspeed.comm.get_rank() == 0:
                    cls._sync_target_model(model, target_model, alpha)

        cls._sync_target_model(model, target_model, alpha)

    def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs) -> None:
        accelerator = kwargs.get('accelerator', None)
        ref_model = kwargs.get('ref_model', None)
        model = kwargs.get('model', None)

        is_zero3_enabled = args.deepspeed_plugin.zero_stage == 3 if args.deepspeed_plugin is not None else False

        if ref_model is not None and state.global_step % self._sync_ref_settings.sync_steps == 0:
            if accelerator is not None:
                unwrapped_model = accelerator.unwrap_model(model)
                self.sync_target_model(
                    model=unwrapped_model,
                    target_model=ref_model,
                    alpha=self._sync_ref_settings.alpha,
                    is_zero3_enabled=is_zero3_enabled,
                )
            else:
                self.sync_target_model(model=model, target_model=ref_model, alpha=self._sync_ref_settings.alpha)



from enum import Enum
from pathlib import Path

from pydantic import model_validator

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class DatasetType(str, Enum):
    CHAT = 'chat'
    PAIR_PREFERENCES = 'pair_preferences'
    SAMPLING = 'sampling'
    DDPO = 'ddpo'
    CLASSIFICATION = 'classification'
    MULTIMODAL = 'multimodal'
    KTO = 'kto'


class DatasetStrategy(str, Enum):
    TRAIN = 'train'
    INFERENCE = 'inference'


class BaseDatasetSettings(ExtraFieldsNotAllowedBaseModel):
    dataset_type: DatasetType


class DatasetSourceSettings(ExtraFieldsNotAllowedBaseModel):
    name: str
    system_prompt: str | None = None
    sample_rate: float | None = None
    num_samples: int | None = None
    records_path: Path | None = None
    records_data: list | None = None
    offset: int | None = None
    n_rows: int | None = None

    @model_validator(mode='after')
    def correct_dataset_sampling_values(self) -> 'DatasetSourceSettings':
        if self.sample_rate is None and self.num_samples is None:
            raise ValueError('neither sample_rate nor num_samples are not set')
        if self.sample_rate is not None and self.num_samples is not None:
            raise ValueError('both sample_rate and num_samples are set')
        if self.offset is not None and self.n_rows is None:
            raise ValueError('both offset and num_rows should be set')
        if self.n_rows is not None and self.offset is None:
            raise ValueError('both offset and num_rows should be set')

        return self


class MultiDatasetSettings(ExtraFieldsNotAllowedBaseModel):
    sources: list[DatasetSourceSettings]
    dataset_type: DatasetType



from typing import Callable

from torch.utils.data import Dataset
from transformers import PreTrainedModel, PreTrainedTokenizerBase, TrainingArguments
from transformers.data.data_collator import (
    DataCollatorForTokenClassification,
    DataCollatorMixin,
)

from turbo_alignment.cherry_picks.chat import ChatCherryPickCallback
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.constants import TRAINER_LOGS_FOLDER
from turbo_alignment.dataset.chat import InferenceChatDataset
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import MetricSettingsRegistry
from turbo_alignment.pipelines.train.base import BaseTrainStrategy
from turbo_alignment.settings.datasets import DatasetStrategy
from turbo_alignment.settings.pipelines.train.sft import SftTrainExperimentSettings
from turbo_alignment.trainers.multigpu import MultiGPUCherryPicksTrainer

logger = get_project_logger()


class TrainSFTStrategy(BaseTrainStrategy[SftTrainExperimentSettings]):
    @staticmethod
    def _get_data_collator(
        experiment_settings: SftTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> Callable:
        return DataCollatorForTokenClassification(tokenizer, pad_to_multiple_of=8)

    @staticmethod
    def _get_cherry_pick_callback(
        experiment_settings: SftTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> ChatCherryPickCallback:
        cherry_pick_settings = experiment_settings.cherry_pick_settings

        cherry_pick_datasets = DatasetLoader[InferenceChatDataset](InferenceChatDataset).load_datasets(
            cherry_pick_settings.dataset_settings, tokenizer=tokenizer, strategy=DatasetStrategy.INFERENCE
        )

        metrics = [
            Metric.by_name(metric.type)(MetricSettingsRegistry.by_name(metric.type)(**metric.parameters))
            for metric in cherry_pick_settings.metric_settings
        ]

        return ChatCherryPickCallback(
            cherry_pick_settings=cherry_pick_settings,
            datasets=cherry_pick_datasets,
            metrics=metrics,
        )

    @staticmethod
    def _get_training_args(experiment_settings: SftTrainExperimentSettings) -> TrainingArguments:
        return TrainingArguments(
            output_dir=str(experiment_settings.log_path / TRAINER_LOGS_FOLDER),
            **experiment_settings.trainer_settings.dict(),
        )

    @staticmethod
    def _get_trainer(
        training_args: TrainingArguments,
        experiment_settings: SftTrainExperimentSettings,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        train_dataset: Dataset,
        val_dataset: Dataset,
        data_collator: DataCollatorMixin,
        **_kwargs,
    ) -> MultiGPUCherryPicksTrainer:
        model.config.use_cache = not experiment_settings.trainer_settings.gradient_checkpointing

        return MultiGPUCherryPicksTrainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            callbacks=[],
            data_collator=data_collator,
            tokenizer=tokenizer,
        )

    def _dataset_and_collator_sanity_check(self, dataset: Dataset, collator: DataCollatorMixin) -> None:
        logger.info(f'Train sample example:\n{dataset[0]}')
        logger.info('Example text check: {example}'.format(example=self.tokenizer.decode(dataset[0]['input_ids'])))
        labels_ids = dataset[0]['labels'].clone()
        labels_ids[labels_ids == -100] = 0
        logger.info('Example label text check {example}'.format(example=self.tokenizer.decode(labels_ids)))

        logger.info(
            'Input ids check: {input_ids}'.format(input_ids=collator([dataset[0], dataset[1]])['input_ids'][0])
        )
        logger.info('Mask check: {mask}'.format(mask=collator([dataset[0], dataset[1]])['attention_mask'][0]))
        logger.info('Labels check: {labels}'.format(labels=collator([dataset[0], dataset[1]])['labels'][0]))



import random
from copy import deepcopy
from pathlib import Path
from typing import Any, overload

from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.data.io import read_jsonl
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.dataset.base import AlignmentDataset
from turbo_alignment.dataset.chat import ChatDatasetRecord, TrainChatDataset
from turbo_alignment.dataset.kto.models import KTODatasetRecord
from turbo_alignment.dataset.registry import KTODatasetTypeRegistry
from turbo_alignment.settings.datasets.base import (
    DatasetSourceSettings,
    DatasetStrategy,
)
from turbo_alignment.settings.datasets.kto import KTODatasetSettings

logger = get_project_logger()


@KTODatasetTypeRegistry.register(DatasetStrategy.TRAIN)
class KTODataset(AlignmentDataset[KTODatasetRecord]):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: KTODatasetSettings,
        tokenizer: PreTrainedTokenizerBase,
    ):
        self._chat_dataset = TrainChatDataset(
            source=source,
            settings=settings.chat_settings,
            tokenizer=tokenizer,
            read=False,
        )
        super().__init__(source=source, settings=settings, tokenizer=tokenizer)

        self._read()

    def convert_records(self, records: list[KTODatasetRecord]) -> list[dict[str, Any] | None]:
        shuffled_records = deepcopy(records)
        random.shuffle(shuffled_records)

        kl_records = []
        chat_records = []

        for record, shuffled_record in zip(records, shuffled_records):
            chat_records.append(ChatDatasetRecord(id=record.id, messages=record.context + [record.answer]))
            kl_records.append(ChatDatasetRecord(id=record.id, messages=record.context + [shuffled_record.answer]))

        tokenized_chat_records = self._chat_dataset.convert_records(chat_records)
        tokenized_kl_records = self._chat_dataset.convert_records(kl_records)

        output: list[dict[str, Any] | None] = []
        for record, tokenized_record, tokenized_kl_record in zip(
            records, tokenized_chat_records, tokenized_kl_records
        ):
            if not (tokenized_record and tokenized_kl_record):
                continue

            chat_tokens = {k: v.squeeze(0) for k, v in tokenized_record.items()}
            kl_tokens = {k: v.squeeze(0) for k, v in tokenized_kl_record.items()}

            output.append(
                {
                    'is_desirable': record.is_desirable,
                    'chat': chat_tokens,
                    'KL': kl_tokens,
                }
            )

        return output

    @staticmethod
    @overload
    def _read_records(records: Path) -> list[KTODatasetRecord]:
        ...

    @staticmethod
    @overload
    def _read_records(records: list[dict]) -> list[KTODatasetRecord]:
        ...

    @staticmethod
    def _read_records(records) -> list[KTODatasetRecord]:
        if isinstance(records, Path):
            return [KTODatasetRecord(**record) for record in read_jsonl(records)]
        if isinstance(records, list):
            return [KTODatasetRecord(**record) for record in records]
        raise NotImplementedError



from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class ModelTransformersSettings(ExtraFieldsNotAllowedBaseModel):
    load_in_8bit: bool = False
    low_cpu_mem_usage: bool = False
    omit_base_model_save: bool | None = None

    trust_remote_code: bool = False



from turbo_alignment.dataset.base.models import DatasetRecord
from turbo_alignment.dataset.chat.models import ChatMessage


class PairPreferenceRecord(DatasetRecord):
    """
    answer_w - preferred answer
    answer_l - other answer
    """

    context: list[ChatMessage]
    answer_w: ChatMessage
    answer_l: ChatMessage
    precomputed_margin: float | None = None



from typing import Generator

from accelerate import Accelerator
from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.tf.loaders import load_model, load_tokenizer
from turbo_alignment.generators.classification import ClassificationGenerator
from turbo_alignment.pipelines.inference.base import BaseInferenceStrategy
from turbo_alignment.settings.pipelines.inference.base import (
    InferenceExperimentSettings,
)


class ClassificationInferenceStrategy(BaseInferenceStrategy[InferenceExperimentSettings]):
    def _get_single_inference_settings(
        self, experiment_settings: InferenceExperimentSettings, accelerator: Accelerator
    ) -> Generator[tuple[PreTrainedTokenizerBase, ClassificationGenerator, str, dict], None, None]:
        save_file_id = 0

        for model_inference_settings in experiment_settings.inference_settings:
            tokenizer = load_tokenizer(
                model_inference_settings.tokenizer_settings,
                model_inference_settings.model_settings,
            )
            model = load_model(model_inference_settings.model_settings, tokenizer)
            model.eval()

            generator = ClassificationGenerator(
                tokenizer=tokenizer,
                model=model,
                accelerator=accelerator,
                batch=model_inference_settings.batch,
            )
            parameters_to_save = {'model_settings': model_inference_settings.model_settings.dict()}

            yield tokenizer, generator, f'single_inference_{save_file_id}.jsonl', parameters_to_save



from pathlib import Path

from turbo_alignment.settings.cherry_pick import MultimodalCherryPickSettings
from turbo_alignment.settings.datasets.multimodal import MultimodalMultiDatasetSettings
from turbo_alignment.settings.modality import (
    Modality,
    ModalityEncoderSettings,
    ModalityProjectorType,
)
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings


class MultimodalTrainExperimentSettings(BaseTrainExperimentSettings):
    modality_encoder_settings_mapping: dict[Modality, ModalityEncoderSettings | None]
    modality_projector_mapping: dict[Modality, ModalityProjectorType | None]
    modality_projector_initialization_mapping: dict[Modality, Path | None]

    train_dataset_settings: MultimodalMultiDatasetSettings
    val_dataset_settings: MultimodalMultiDatasetSettings

    cherry_pick_settings: MultimodalCherryPickSettings



from enum import Enum
from typing import Any, Callable

import numpy as np

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class MetricType(str, Enum):
    ACCURACY: str = 'accuracy'
    ROUGE: str = 'rouge'
    METEOR: str = 'meteor'
    REWARD: str = 'reward'
    DIVERSITY: str = 'diversity'
    PERPLEXITY: str = 'perplexity'
    SELF_BLEU: str = 'self_bleu'
    LENGTH: str = 'length'
    DIST_N: str = 'dist_n'
    KL: str = 'kl'
    TOOL_CALL_METRICS: str = 'tool_call_metrics'
    RETRIEVAL_UTILITY: str = 'retrieval_utility'


class ElementWiseScores(ExtraFieldsNotAllowedBaseModel):
    label: str
    values: list[Any]
    average_function: Callable[[list[Any]], float] = np.mean


class MetricSettings(ExtraFieldsNotAllowedBaseModel):
    type: MetricType
    parameters: dict[str, Any]


class MetricResults(ExtraFieldsNotAllowedBaseModel):
    element_wise_scores: list[ElementWiseScores]
    need_average: bool = False



from typing import Annotated

from pydantic import Field

from turbo_alignment.dataset.base.models import DatasetRecord
from turbo_alignment.dataset.multimodal.models import (
    MultimodalAudioMessage,
    MultimodalImageMessage,
    MultimodalTextMessage,
)
from turbo_alignment.settings.generators.outputs.base import BaseInferenceOutput
from turbo_alignment.settings.generators.outputs.chat import AnswerMessage


class MultimodalInferenceOutput(BaseInferenceOutput, DatasetRecord):
    messages: list[
        Annotated[MultimodalImageMessage | MultimodalAudioMessage | MultimodalTextMessage, Field(discriminator='type')]
    ]
    label: str | None = None
    answers: list[AnswerMessage]



from pathlib import Path

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.model import PreTrainedModelSettings
from turbo_alignment.settings.tf.tokenizer import TokenizerSettings


class MergeAdaptersToBaseModelSettings(ExtraFieldsNotAllowedBaseModel):
    model_settings: PreTrainedModelSettings
    tokenizer_settings: TokenizerSettings

    adapter_path: Path
    save_path: Path

    max_shard_size: str = '400MB'
